char sol[] =
"stage 1 curchild -1 type sum\n\
stage 2 curchild 0 type sum\n\
stage 1 curchild -1 type var\n\
stage 8 curchild -1 type var\n\
stage 4 curchild 0 type sum\n\
stage 2 curchild 1 type sum\n\
stage 1 curchild -1 type prod\n\
stage 2 curchild 0 type prod\n\
stage 1 curchild -1 type var\n\
stage 8 curchild -1 type var\n\
stage 4 curchild 0 type prod\n\
stage 2 curchild 1 type prod\n\
stage 1 curchild -1 type var\n\
stage 8 curchild -1 type var\n\
stage 4 curchild 1 type prod\n\
stage 2 curchild 2 type prod\n\
stage 1 curchild -1 type val\n\
stage 8 curchild -1 type val\n\
stage 4 curchild 2 type prod\n\
stage 8 curchild -1 type prod\n\
stage 4 curchild 1 type sum\n\
stage 8 curchild -1 type sum\n\
";
